@@ -12,10 +12,17 @@ _IS_QISKIT_AVAILABLE = True
 try:
     from qiskit.circuit.quantumcircuit import QuantumCircuit
     from qiskit.qobj.qasm_qobj import QasmQobjExperiment
+    from qiskit.quantum_info.operators.symplectic.clifford import Clifford
     from .util import convert_qiskit_circuit_to_qasm_experiment
 except ImportError:
     _IS_QISKIT_AVAILABLE = False
 
+_IS_NUMPY_AVAILABLE = True
+try:
+    import numpy as np
+except:
+    _IS_NUMPY_AVAILABLE = False
+
 
 class QrackSimulator:
     """Interface for all the QRack functionality.
@@ -2364,6 +2371,151 @@ class QrackSimulator:
         Qrack.qrack_lib.SetTInjection(self.sid, iti)
         self._throw_if_error()
 
+    def set_hardware_encoded(self, she):
+        """Set option to encode t-gadgets for hardware
+
+        If t-injection is on, and we output a stabilizer state to file,
+        this option turns on auxiliary error correction channels, if true,
+        allowing ancilla syndrome correction with mid-ciruit measurement
+        instead of post selection.
+        Note that this encoding option is off by default.
+
+        Args:
+            she: use "stabilizer hardware encoding"
+
+        Raises:
+            RuntimeError: QrackSimulator raised an exception.
+        """
+        Qrack.qrack_lib.SetStabilizerHardwareEncoded(self.sid, she)
+        self._throw_if_error()
+
+    def out_to_file(self, filename):
+        """Output state to file (stabilizer only!)
+
+        Outputs the hybrid stabilizer state to file.
+
+        Args:
+            filename: Name of file
+        """
+        Qrack.qrack_lib.qstabilizer_out_to_file(self.sid, filename.encode('utf-8'))
+        self._throw_if_error()
+
+    def in_from_file(filename, is_binary_decision_tree = False, is_paged = True, is_cpu_gpu_hybrid = True, is_opencl = True, is_host_pointer = False):
+        """Input state from file (stabilizer only!)
+
+        Reads in a hybrid stabilizer state from file.
+
+        Args:
+            filename: Name of file
+        """
+        out = QrackSimulator(
+            qubitCount=1,
+            isSchmidtDecomposeMulti=False,
+            isSchmidtDecompose=False,
+            isStabilizerHybrid=True,
+            isBinaryDecisionTree=is_binary_decision_tree,
+            isPaged=is_paged,
+            isCpuGpuHybrid=is_cpu_gpu_hybrid,
+            isOpenCL=is_opencl,
+            isHostPointer=is_host_pointer
+        )
+        Qrack.qrack_lib.qstabilizer_in_from_file(out.sid, filename.encode('utf-8'))
+        out._throw_if_error()
+
+        return out
+
+    def file_to_qiskit_circuit(filename, is_hardware_encoded=False):
+        """Convert an output state file to a Qiskit circuit
+
+        Reads in an (optimized) circuit from a file named
+        according to the "filename" parameter and outputs
+        a Qiskit circuit.
+
+        Args:
+            filename: Name of file
+
+        Raises:
+            RuntimeErorr: Before trying to file_to_qiskit_circuit() with
+                QrackCircuit, you must install Qiskit, numpy, and math!
+        """
+        if not (_IS_QISKIT_AVAILABLE and _IS_NUMPY_AVAILABLE):
+            raise RuntimeError(
+                "Before trying to file_to_qiskit_circuit() with QrackCircuit, you must install Qiskit, numpy, and math!"
+            )
+
+        lines = []
+        with open(filename, 'r') as file:
+            lines = file.readlines()
+
+        logical_qubits = int(lines[0])
+        stabilizer_qubits = int(lines[1])
+        rows = stabilizer_qubits << 1
+
+        tableau = []
+        for line in lines[2:(rows + 2)]:
+            bits = line.split()
+            row = []
+            for bit in bits:
+                row.append(bool(int(bit)))
+            row[-1] = (int(bits[-1]) >> 1) & 1
+            tableau.append(row)
+
+        non_clifford_gates = []
+        g = 0
+        for line in lines[(rows + 2):]:
+            i = 0
+            tokens = line.split()
+            op = np.zeros((2,2), dtype=complex)
+            row = []
+            for _ in range(2):
+                amp = tokens[i].replace("(","").replace(")","").split(',')
+                row.append(float(amp[0]) + float(amp[1])*1j)
+                i = i + 1
+            l = math.sqrt(np.real(row[0] * np.conj(row[0]) + row[1] * np.conj(row[1])))
+            op[0][0] = row[0] / l
+            op[0][1] = row[1] / l
+
+            if np.abs(op[0][0] - row[0]) > 1e-5:
+                print("Warning: gate ", str(g), " might not be unitary!")
+            if np.abs(op[0][1] - row[1]) > 1e-5:
+                print("Warning: gate ", str(g), " might not be unitary!")
+
+            row = []
+            for _ in range(2):
+                amp = tokens[i].replace("(","").replace(")","").split(',')
+                row.append(float(amp[0]) + float(amp[1])*1j)
+                i = i + 1
+            l = math.sqrt(np.real(row[0] * np.conj(row[0]) + row[1] * np.conj(row[1])))
+            op[1][0] = row[0] / l
+            op[1][1] = row[1] / l
+
+            ph = np.real(np.log(np.linalg.det(op)) / 1j)
+
+            op[1][0] = -np.exp(1j * ph) * np.conj(op[0][1])
+            op[1][1] = np.exp(1j * ph) * np.conj(op[0][0])
+
+            if np.abs(op[1][0] - row[0]) > 1e-5:
+                print("Warning: gate ", str(g), " might not be unitary!")
+            if np.abs(op[1][1] - row[1]) > 1e-5:
+                print("Warning: gate ", str(g), " might not be unitary!")
+
+            non_clifford_gates.append(op)
+            g = g + 1
+
+        clifford = Clifford(tableau)
+
+        circ = clifford.to_circuit()
+        for i in range(len(non_clifford_gates)):
+            circ.unitary(non_clifford_gates[i], [i])
+
+        if is_hardware_encoded:
+            for i in range(logical_qubits, stabilizer_qubits, 2):
+                circ.h(i + 1)
+                circ.cz(i, i + 1)
+                circ.h(i + 1)
+
+        return circ
+
     def _apply_pyzx_op(self, gate):
         if gate.name == "XPhase":
             self.r(Pauli.PauliX, math.pi * gate.phase, gate.target)
@@ -2455,6 +2607,8 @@ class QrackSimulator:
                 float(operation.params[1]),
                 float(operation.params[2]),
             )
+        elif (name == 'unitary') and (len(operation.qubits) == 1):
+            self._sim.mtrx(operation.params[0].flatten(), operation.qubits[0])
         elif name == 'r':
             self._sim.u(
                 operation.qubits[0],
